/*
	Function combination function. Takes up to two
	functions, each of which take up to one argument,
	and combines them into a single function that takes
	one argument.
*/
var PrivateWaiting, PublicWaiting, WaitGroup, buildPublicWaiting, cancelFn, combine, finishFn;
var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __slice = Array.prototype.slice;
combine = function(fn1, fn2) {
  if (fn1 && fn2) {
    return function(arg) {
      fn1(arg);
      fn2(arg);
    };
  } else if (fn1) {
    return fn1;
  } else {
    return fn2;
  }
};
/* 
	Helper functions to build the callbacks that get
	passed to promised functions
*/
finishFn = function(waiting) {
  return function(data) {
    waiting.finish(data);
  };
};
cancelFn = function(waiting) {
  return function(err) {
    waiting.cancel(err);
  };
};
/* 
	Private Waiting class -- wrapped in the public
	version of the class and returned by flow.wait()
*/
PrivateWaiting = (function() {
  function PrivateWaiting(wrappedDependencies, callback) {
    var dependency, isReady, _fn, _i, _len;
    this.callback = callback;
    this.err = "Error.";
    this.isFinished = this.isCancelled = false;
    this.dependencies = [];
    this.onCancel = this.onFinish = null;
    isReady = true;
    _fn = __bind(function(dependency) {
      isReady = dependency.isFinished() && isReady && !dependency.isCancelled();
      if (!dependency.isFinished()) {
        this.dependencies.push(dependency);
        dependency.finished(__bind(function(data) {
          return this.notify(dependency);
        }, this));
        return dependency.cancelled(__bind(function(err) {
          return this.cancel(err);
        }, this));
      }
    }, this);
    for (_i = 0, _len = wrappedDependencies.length; _i < _len; _i++) {
      dependency = wrappedDependencies[_i];
      _fn(dependency);
    }
    if (isReady) {
      this.callDeferred(this.callback);
    }
  }
  PrivateWaiting.prototype.callDeferred = function(fn) {
    if (!this.isCancelled && fn) {
      return fn(finishFn(this), cancelFn(this));
    }
  };
  PrivateWaiting.prototype.finished = function(fn) {
    if (this.isFinished) {
      return fn(this.data);
    } else {
      return this.onFinish = combine(this.onFinish, fn);
    }
  };
  PrivateWaiting.prototype.finish = function(data) {
    if (this.isCancelled) {
      return;
    }
    this.data = data;
    this.isFinished = true;
    if (this.onFinish) {
      return this.onFinish(this.data);
    }
  };
  PrivateWaiting.prototype.cancelled = function(fn) {
    if (this.isCancelled) {
      return fn(this.err);
    } else {
      return this.onCancel = combine(this.onCancel, fn);
    }
  };
  PrivateWaiting.prototype.cancel = function(err) {
    this.err = err;
    this.isCancelled = true;
    if (this.onCancel) {
      this.onCancel(err);
    }
    return this.dependencies = [];
  };
  PrivateWaiting.prototype.notify = function(readyDependency) {
    var index;
    if (this.isCancelled || this.isFinished) {
      return;
    }
    index = this.dependencies.indexOf(readyDependency);
    if (index >= 0) {
      this.dependencies.splice(index, 1);
      if (this.dependencies.length <= 0) {
        return this.callDeferred(this.callback);
      }
    }
  };
  return PrivateWaiting;
})();
/*
	Public Waiting class -- wraps around private
	version, and only exposes certain functions.
*/
PublicWaiting = (function() {
  function PublicWaiting(dependencies, callback) {
    var wrapped;
    wrapped = new PrivateWaiting(dependencies, callback);
    this.cancelled = function(fn) {
      wrapped.cancelled(fn);
      return this;
    };
    this.finished = function(fn) {
      wrapped.finished(fn);
      return this;
    };
    this.isFinished = function() {
      return wrapped.isFinished;
    };
    this.isCancelled = function() {
      return wrapped.isCancelled;
    };
  }
  return PublicWaiting;
})();
/*
	Helper function to build PublicWaiting objects
	from optional arguments.
*/
buildPublicWaiting = function(argsArray) {
  var arg, autoFinish, lastArg;
  autoFinish = function(finish, cancel) {
    return finish();
  };
  if (argsArray.length === 0) {
    return new PublicWaiting([], autoFinish);
  }
  if (argsArray.length === 1) {
    arg = argsArray[0];
    if (typeof arg === 'function') {
      return new PublicWaiting([], arg);
    }
    return new PublicWaiting([arg], autoFinish);
  }
  lastArg = argsArray[argsArray.length - 1];
  if (typeof lastArg === 'function') {
    return new PublicWaiting(argsArray.slice(0, -1), lastArg);
  }
  return new PublicWaiting(argsArray, autoFinish);
};
/*
	Wait Group class -- holds multiple wait objects
	dependencies, and allows dynamic creation of new
	wait objects that extend those groups.
*/
WaitGroup = (function() {
  function WaitGroup(dependencies) {
    this.dependencies = dependencies;
  }
  WaitGroup.prototype.wait = function() {
    var args;
    args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    return buildPublicWaiting(this.dependencies.concat(args));
  };
  return WaitGroup;
})();
/*
	Public creation functions
*/
exports.wait = function() {
  var args;
  args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
  return buildPublicWaiting(args);
};
exports.group = function() {
  var dependencies;
  dependencies = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
  return new WaitGroup(dependencies);
};