vows = require 'vows'
assert = require 'assert'
events = require 'events'
flow = require '../index'

exports.suite = suite = vows.describe('Heavy Flow')

suite.addBatch
	'Wait objects':
		topic: () -> flow.wait

		'that are empty':
			topic: (topic) -> topic

			'should finish immediately after being created': (topic) ->
				test = topic()
				assert.isTrue test.isFinished()
			'should not be cancelled immediately after being created': (topic) ->
				test = topic()
				assert.isFalse test.isCancelled()
		'that have no dependencies':
			topic: (topic) -> topic
			'should run immediately': (topic) ->
				ran = false
				test = topic (finish, cancel) ->
					ran = true
					finish()
				assert.isTrue test.isFinished() && ran
			'should not be cancelled after finishing successfully': (topic) ->
				test = topic (finish, cancel) ->
					finish()
				assert.isFalse test.isCancelled()
			'should be cancelled after cancelling': (topic) ->
				test = topic (finish, cancel) ->
					cancel()
				assert.isTrue test.isCancelled()
			'should not be finished after cancelling': (topic) ->
				test = topic (finish, cancel) ->
					cancel()
				assert.isFalse test.isFinished()
						
			'and have callbacks':
				topic: (topic) -> topic
				'should call finished callbacks after finishing': (topic) ->
					called = false
					test = topic (finish, cancel) ->
						finish()
					.finished () ->
						called = true
					assert.isTrue called
				'should call cancelled callbacks after cancelling': (topic) ->
					called = false
					test = topic (finish, cancel) ->
						cancel()
					.cancelled () ->
						called = true
					assert.isTrue called
				'should not call finished callbacks after cancelling': (topic) ->
					called = false
					test = topic (finish, cancel) ->
						cancel()
					.finished () ->
						called = true
					assert.isFalse called
				'should not call cancelled callbacks after finishing': (topic) ->
					called = false
					test = topic (finish, cancel) ->
						finish()
					.cancelled () ->
						called = true
					assert.isFalse called

				'should pass data to finished functions': (topic) ->
					data = null
					test = topic (finish, cancel) ->
						finish { dumbledore: true }
					.finished (result) ->
						data = result
					assert.isTrue data.dumbledore
				'should pass error messages to cancelled functions': (topic) ->
					errMsg = ''
					test = topic (finish, cancel) ->
						cancel 'err'
					.cancelled (err) ->
						errMsg = err
					assert.equal errMsg, 'err'
		
		'that have synchronous dependencies':
			topic: (topic) -> topic

			'should run in order, if none cancel': (topic) ->
				firstRan = false
				first = topic (finish, cancel) ->
					firstRan = true
					finish()
				second = topic first, (finish, cancel) ->
					if firstRan then finish() else cancel()
				assert.isTrue second.isFinished()

			'should not run if their only dependency cancels': (topic) ->
				first = topic (finish, cancel) ->
					cancel()
				second = topic first, (finish, cancel) ->
					finish()
				assert.isFalse second.isFinished()

			'should not run if any of their multiple dependencies cancel': (topic) ->
				first = topic (finish, cancel) ->
					cancel()
				second = topic (finish, cancel) ->
					finish()
				third = topic first, second, (finish, cancel) ->
					finish()
				assert.isFalse third.isFinished()

			'should run their finished callbacks before calling their dependents': (topic) ->
				finishedRan = false
				first = topic().finished () ->
					finishedRan = true
				second = topic (finish, cancel) ->
					if finishedRan then finish() else cancel()
				assert.isTrue second.isFinished()

			'should run their cancellation callbacks before cancelling their dependents': (topic) ->
				cancelledRan = false
				first = topic (finish, cancel) -> cancel()
				first.cancelled (err) ->
					cancelledRan = true
				second = topic().cancelled (err) ->
					assert.isTrue cancelledRan

			'should pass cancellation errors to their dependants': (topic) ->
				first = topic (finish, cancel) ->
					cancel('test')
				errMsg = ''
				second = topic first, (finish, cancel) ->
					finish()
				.cancelled (msg) ->
					errMsg = msg
				assert.equal errMsg, 'test'

			'should allow optional ready function arguments':
				topic: (topic) -> topic

				'should run if their dependency finishes': (topic) ->
					first = topic()
					second = topic first
					assert.isTrue second.isFinished()

				'should cancel if their dependency cancels': (topic) ->
					first = topic (finish, cancel) ->
						cancel()
					second = topic first
					assert.isTrue second.isCancelled()

				'should get the correct error messages if their dependency cancels': (topic) ->
					first = topic (finish, cancel) ->
						cancel('test')
					errMsg = ''
					second = topic(first).cancelled (msg) ->
						errMsg = msg
					assert.equal errMsg, 'test'
		
		'that have asynchronous dependencies':
			topic: (topic) -> topic

			'that finish':
				topic: (topic) ->
					firstCalled = false
					first = topic (finish, cancel) ->
						setTimeout finish, 500
					.finished () ->
						firstCalled = true
					second = topic first, (finish, cancel) =>
						if firstCalled then this.callback(null, 'success') else this.callback('err', null)
					return
				'should run in order': (err, data) ->
					assert.equal err, null
			
			'that cancel':
				topic: (topic) ->
					first = topic (finish, cancel) ->
						setTimeout cancel, 500
					second = topic first, (finish, cancel) =>
						this.callback 'err', null
						finish()
					.cancelled () =>
						this.callback null, 'success'
					return
				'should get cancelled': (err, data) ->
					assert.equal err, null

			'that are asynchronous':
				topic: (topic) -> topic

				'and have finished functions':
					topic: (topic) ->
						test = topic (finish, cancel) ->
							setTimeout finish, 500
						.finished () =>
							this.callback null, 'success'
						return
					'should finish': (err, data) ->
						assert.equal err, null

				'and have cancelled callbacks, and themselves cancel':
					topic: (topic) ->
						test = topic (finish, cancel) ->
							setTimeout cancel, 500
						.cancelled () =>
							this.callback null, 'success'
						return
					'should call the callbacks': (err, data) ->
						assert.equal err, null

				'and have cancelled callbacks but never cancel':
					topic: (topic) ->
						test = topic (finish, cancel) ->
							setTimeout finish, 500
						.cancelled () =>
							this.callback 'err', null
						.finished () =>
							this.callback null, 'success'
						return
					'should not call the callbacks': (err, data) ->
						assert.equal err, null

				'and have finished callbacks but never finish':
					topic: (topic) ->
						test = topic (finish, cancel) ->
							setTimeout cancel, 500
						.finished () =>
							this.callback 'err', null
						.cancelled () =>
							this.callback null, 'success'
						return
					'should not call the callbacks': (err, data) ->
						assert.equal err, null
				
				'and have finished callbacks added asynchronously':
					topic: (topic) ->
						test = topic (finish, cancel) =>
							addCallback = () =>
								test.finished (data) =>
									this.callback null, data
							f = () ->
								finish { dumbledore: true }
								setTimeout addCallback, 250
								
							setTimeout f, 250
						return

					'should call the callbacks': (err, data) ->
						assert.notEqual data, null
					'should pass the correct data': (err, data) ->
						assert.isTrue data.dumbledore

				'and have cancelled callbacks added asynchronously':
					topic: (topic) ->
						test = topic (finish, cancel) =>
							addCallback = () =>
								test.cancelled (err) =>
									this.callback err, null
							c = () ->
								cancel 'err'
								setTimeout addCallback, 250
							setTimeout c, 250
						return

					'should call the callbacks': (err, success) ->
						assert.notEqual err, null
					'should pass the correct message': (err, success) ->
						assert.equal err, 'err'
